// Copyright (c) Tech Soft 3D, Inc.
//
// The information contained herein is confidential and proprietary to Tech Soft 3D, Inc.,
// and considered a trade secret as defined under civil and criminal statutes.
// Tech Soft 3D, Inc. shall pursue its civil and criminal remedies in the event of
// unauthorized use or misappropriation of its trade secrets.  Use of this information
// by anyone other than authorized employees of Tech Soft 3D, Inc. is granted only under
// a written non-disclosure agreement, expressly prescribing the scope and manner of such use.

// shell_texture_hsf.cpp
//
// This is intended to be as close as possible to a minimal example that uses textures
// on shells.  The most difficult part of the process of writing shells is to make sure
// that dependencies are resolved.  TK_Color's use defined TK_Texture's which require
// TK_Image's, all of which must appear in the right order in the file.  
//
// This example intentionally uses a variety of ways to fill in the data of opcode handlers.
//
// For the sake of simplicity, we break a few of the rules that we would normally
// advocate for better performance, such as reusing opcode handlers.
//
// In the process of creating this demo, we discovered several settings that could be made
// implicit.  Such settings are marked with a "xxx" comment.  HOOPS/stream has been modified
// to either make them the default or automatically performed as a result of other actions.
// 

#include "BStream.h"
#include "BOpcodeShell.h"
#include <stdio.h>

#define WIDTH 128
#define HEIGHT 128
#define countof(x) (int)(sizeof(x)/sizeof(x[0]))


// export the HSF object to a file
static void WriteObject(BBaseOpcodeHandler * handler, BStreamFileToolkit * tk, FILE * fp,
                        char * buffer, int size) 
{
    TK_Status status;

    // ask the toolkit to write out the object; the toolkit will write out as much as
    // the current buffer can hold
    status = handler->Write(*tk);

    // if there is still more data left to write; we have to loop through 
    // generating and writing buffers of data until we're done
    while (status == TK_Pending) 
    {
        // due to compression, preparing a new buffer might immediately fill that buffer
        // so we need an extra check to possibly export that buffer and prepare a new buffer
        while (status == TK_Pending) 
        {
            // ask the toolkit how much of the buffer is currently filled
            // this will normally be the same as the buffer size that we passed in
            int count = tk->CurrentBufferLength();

            // write the current out to our file since this example deals with 
            // exporting a file; however, instead of writing to a file, the buffer of
            // data could be sent to an application data-structure, or transmitted
            // over a network ('streamed'), etc...
            fwrite(buffer, sizeof (char), count, fp);

            // tell the toolkit where it can resume writing data; in this case, we
            // can reuse the same buffer once we're done with it
            status = tk->PrepareBuffer(buffer, size);
        }

        // ask the toolkit to continue trying to write out the object's data
        status = handler->Write(*tk);
    }
}


class TK_My_Shell: public TK_Shell
{
public:
    TK_My_Shell():TK_Shell() { };
    ~TK_My_Shell(){ Reset(); };

    /* Normally you would use the given key to go look up this shell from some
     * graphics database or another (possibly but not necessarily HOOPS/3dgs).
     * for the purposes of this demo, however, we just create the data on the spot
     * with hard-code. */
    TK_Status Interpret (BStreamFileToolkit & tk, ID_Key key, int lod) alter 
    {
        TK_Status status = TK_Normal;
        const float points[8][3] = {
            {    0,    0, 0 },
            {    0,    1, 0 },
            {    1,    1, 0 },
            {    1,    0, 0 },
            { 0.2f, 0.2f, 0 },
            { 0.2f, 0.8f, 0 },
            { 0.8f, 0.8f, 0 },
            { 0.8f, 0.2f, 0 }
        };
        const int faces[] = { 4, 0, 1, 2, 3, -4, 4, 5, 6, 7 };
        int index;
        bool is_firstpass = tk.KeyToIndex( key, index ) == TK_NotFound;

        InitSubop( tk, lod, is_firstpass );
        SetSubop( GetSubop() & ~TKSH_TRISTRIPS );

        //allocate and initialize the vertices
        SetPoints( countof(points), (const float *)points );
        //allocate and initialize the face indices
        SetFaces( countof(faces), faces );
        //simply reuse the world space coordinates as texture coordinates
        SetVertexParameters( (const float *)points ); 
        return status;
    }

    void Reset() {
        TK_Shell::Reset();
    }
};  // end declaration and definition of TK_My_Shell 


// This rather simplistic function simply creates a checkerboard pattern.  In any 
// real application, this kind of function would load a real image -- either from
// a file or from a graphics database.
void create_checkerboard( TK_Image *image, int width, int height, char *name )
{
    unsigned char *ptr;
    int i, j, black;

	image->SetPosition( 0, 0, 0 );
    image->SetSize( width, height );
    image->SetBytes( 3*width*height );
    ptr = (unsigned char *)image->GetBytes( );
    image->SetFormat( TKO_Image_RGB );
    image->SetFormat( image->GetFormat() | TKO_Image_Is_Named ); //xxx
    image->SetName( name );
    image->SetCompression( TKO_Compression_None ); //xxx
    for( j = 0 ; j < height ; j++ ) {
        for( i = 0 ; i < width ; i++ ) {
            //bitwise XOR the low bit of row/column divided by sixteen
            black = ((i / 16) & 0x1) ^ ((j / 16) & 0x1);
            if( black )
                ptr[0] = ptr[1] = ptr[2] = 128;
            else
                ptr[0] = ptr[1] = ptr[2] = 255;
            ptr += 3;
        }
    }
} // end function create_checkerboard


int main() 
{
    BBaseOpcodeHandler *handler;
    BStreamFileToolkit *mytoolkit = new BStreamFileToolkit;
    char mybuffer[4096];
    int size=4096;
    int count;
    FILE *flptr;

    // open a binary file for writing
    flptr=fopen("testfile4.hsf", "wb");

    // tell the toolkit where to begin writing data; the size of the 'utility' buffer
    // can be anything; for this example, we use a 4k buffer
    mytoolkit->PrepareBuffer(mybuffer, size);

    // create a TK_Header object and export it's associated HSF objects,
    // which include the TKE_Comment and TKE_File_Info objects
    // TK_Header automatically formats the TKE_Comment to contain file version 
    // information.  
    handler = new TK_Header();
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // export TKE_Open_Segment so that the image can be in a subsegment with separate visibility settings
    // There is currently, unfortunately, no way to avoid having the image included in bounding volume
    // calculations if we put it into a subsegment like this.  The alternative, though, is to put the image
    // into a top-level directory, such as "/images" or "?Include Library".  The drawback of that approach,
    // though, is that directories such as those are difficult to clean up when the window is closed.
    handler = new TK_Open_Segment;
    ((TK_Open_Segment *)handler)->SetSegment("images");
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;
    
        // use TK_Visibility to turn the visibility of images off.  This image is solely a resource for
        // textures, not something that is supposed to appear in the scene
        handler = new TK_Visibility;
        ((TK_Visibility *)handler)->SetGeometry( TKO_Geo_Image );
        ((TK_Visibility *)handler)->SetValue( ~TKO_Geo_All_Visibles );
        WriteObject(handler, mytoolkit, flptr, mybuffer, size);
        delete handler;

        // as shown in this example, the function that initializes the data need not necessarily 
        // be called "interpret" 
        handler = new TK_Image;
        create_checkerboard( (TK_Image *) handler, WIDTH, HEIGHT, "checkerboard" );
        WriteObject(handler, mytoolkit, flptr, mybuffer, size);
        delete handler;

    // close the segment
    handler = new TK_Close_Segment;
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // put the model in a different segment
    handler = new TK_Open_Segment;
    ((TK_Open_Segment *)handler)->SetSegment("model");
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

        // turn off polygon handedness since in this example we are creating a flat object that 
        // we want to be visible from both sides.
        handler = new TK_Heuristics;
        ((TK_Heuristics *)handler)->SetMask( TKO_Heuristic_Polygon_Handedness );
        ((TK_Heuristics *)handler)->SetValue( 0 );
        WriteObject(handler, mytoolkit, flptr, mybuffer, size);
        delete handler;


        // LOD 0 (the base resolution) of the shell id'd by key 13
        handler = new TK_My_Shell;
        handler->Interpret(*mytoolkit, 13, 0);
        WriteObject(handler, mytoolkit, flptr, mybuffer, size);
        delete handler;

        // define a texture.  In this example, we simply provide the association between a
        // named image and a named texture resource, but TK_Texture also has settings for things
        // like tiling settings, texture matrices and alternate parameterization sources.
        // The TK_Image is required to have been written at this point
        handler = new TK_Texture;
        ((TK_Texture *)handler)->SetFlags( 0 ); //xxx
        ((TK_Texture *)handler)->SetImage( "checkerboard" );
        ((TK_Texture *)handler)->SetName( "my texture checkerboard" );
        WriteObject(handler, mytoolkit, flptr, mybuffer, size);
        delete handler;

        // export a setting to use the texture as the diffuse color.  The TK_Texture is required to have
        // already been written at this point.  Since the texture uses all default options, however,
        // it would have been legal to use the image as a color directly, a la SetDiffuseName("checkerboard")
        handler = new TK_Color;
        ((TK_Color *)handler)->SetGeometry( TKO_Geo_Face );
        ((TK_Color *)handler)->SetChannels( 1 << TKO_Channel_Diffuse ); //xxx
        ((TK_Color *)handler)->SetDiffuseName( "my texture checkerboard"  );
        WriteObject(handler, mytoolkit, flptr, mybuffer, size);
        delete handler;

    // close the segment
    handler = new TK_Close_Segment;
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // export a TKE_Termination object
    handler = new TK_Terminator(TKE_Termination);
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // ask the toolkit how much of the final buffer is filled
    count = mytoolkit->CurrentBufferLength();

    // write out the final buffer
    fwrite(mybuffer, sizeof (char), count, flptr);

    // close the file
    fclose(flptr);

    // cleanup the toolkit object
    delete mytoolkit;

    return 0;
}

